package me.icymint.libra.sage.core;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import javax.sql.DataSource;

import me.icymint.libra.sage.core.dialect.Dialect;
import me.icymint.libra.sage.core.dialect.SqlDialect;
import me.icymint.libra.sage.core.query.AbstractExecute;
import me.icymint.libra.sage.core.query.AbstractQuery;
import me.icymint.libra.sage.core.query.Execute;
import me.icymint.libra.sage.core.query.Executes;
import me.icymint.libra.sage.core.query.MetaDataQuery;
import me.icymint.libra.sage.core.query.SqlQuery;
import me.icymint.libra.sage.core.result.MetaDataResultHandler;
import me.icymint.libra.sage.core.result.ScalarListResult;

/**
 * 本类提供了Sage项目的Jdbc执行基本方法，以及一些数据库基本信息的获取。 <br/>
 * {@link #initDao(Connection)} 方法提供了自动创建数据库的支持。
 * 
 * @author Daniel
 * @datetime 2012-11-9上午11:48:33
 * 
 */
public abstract class JdbcSupport {

	private class CreateDatabase implements SqlQuery<Integer> {

		private final String database;

		CreateDatabase(String db) {
			this.database = db;

		}

		@Override
		public Integer query(Connection conn) throws SQLException {
			createDatabase(conn, database);
			return 1;
		}

	}

	/**
	 * 初始化本类。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-12上午8:31:21
	 * 
	 */
	private class InitialDao implements SqlQuery<Void> {

		@Override
		public Void query(Connection conn) throws SQLException {
			initDao(conn);
			return null;
		}

	}

	/**
	 * 查询已存在的所有数据库。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-9上午11:59:29
	 * 
	 */

	private class QueryCatalogs extends ScalarListResult<String> implements
			MetaDataResultHandler<List<String>> {

		public QueryCatalogs() {
			super(1);
		}

		@Override
		public ResultSet queryResultSet(DatabaseMetaData metadata)
				throws SQLException {
			return metadata.getCatalogs();
		}

	}

	private QueryCatalogs qc = new QueryCatalogs();

	private MetaDataQuery<List<String>> mdq = new MetaDataQuery<List<String>>(
			qc);

	/**
	 * 执行累计计数。
	 */
	private AtomicInteger ai = new AtomicInteger();

	/**
	 * 数据源。
	 */
	private final DataSource ds;

	public final static String NULL = "1234567890";

	/**
	 * 默认数据库。
	 */
	private String defaultCatalog;
	private SqlDialect dialect = null;

	/**
	 * 该构造器支持DataSource中已经存在数据库信息的情况。
	 * 
	 * @param ds
	 */
	protected JdbcSupport(DataSource ds) {
		this(ds, NULL);
	}

	protected JdbcSupport(DataSource ds, String database) {
		this.ds = ds;
		setDefaultCatalog(database);
		try {
			this.queryCore(new InitialDao());
		} catch (SQLException e) {
			throw new RuntimeException("数据源连接有问题！\n" + ds, e);
		}
	}

	/**
	 * 检查当前数据库设置是否正确，如果不正确则设置。
	 * 
	 * @param conn
	 * @throws SQLException
	 */
	private void checkAndSetCatalog(Connection conn) throws SQLException {
		if (this.getDefaultCatalog() == null) {
			this.defaultCatalog = dialect.getDefaultCatalog(conn);
		}
		String cat = dialect.getCurrentCatalog(conn);
		if (cat == null || !cat.equals(getDefaultCatalog())) {
			try {
				dialect.setCurrentCatalog(conn, getDefaultCatalog());
			} catch (SQLException e) {
				if (dialect.isSupportAutoCreateDatabase()) {
					createDatabase(conn, getDefaultCatalog());
					dialect.setCurrentCatalog(conn, getDefaultCatalog());
				} else
					throw e;
			}
		}
	}

	/**
	 * 关闭数据库。
	 * 
	 * @param conn
	 * @throws SQLException
	 */
	private void closeConnection(Connection conn) throws SQLException {
		try {
			conn.close();
		} catch (SQLException e) {
			System.out.println("已关闭");
		}
	}

	private void createDatabase(Connection conn, String database)
			throws SQLException {
		List<String> cs = mdq.query(conn);
		if (!cs.contains(database)) {
			System.out.println("数据库不存在，故自动创建" + database);
			new Execute(dialect.getCreateDatabase(database)).query(conn);
		}
	}

	public void createDatabase(String database) throws SQLException {
		this.queryCore(new CreateDatabase(database));
	}

	/**
	 * 执行Jdbc执行指令。
	 * 
	 * @param exe
	 * @throws SQLException
	 */
	protected <S> void execute(AbstractExecute<S> exe) throws SQLException {
		this.queryCore(exe);
	}

	public void executeBatchs(String... sql) throws SQLException {
		queryCore(new Executes(sql));
	}

	/**
	 * 获取累计执行次数。 该计数只是统计请求Connection的次数。
	 */
	public int get() {
		return ai.get();
	}

	/**
	 * 获取数据库。
	 * 
	 * @return
	 */
	public List<String> getCatalogs() {
		try {
			return this.queryMetaData(qc);
		} catch (SQLException e) {
			throw new RuntimeException("数据库列表获取出错！", e);
		}
	}

	/**
	 * 获取Connection，并且集成了自动检查的功能。
	 * 
	 * @throws SQLException
	 */
	private Connection getConnection() throws SQLException {
		Connection conn = this.getPureConnection();
		checkAndSetCatalog(conn);
		return conn;
	}

	/**
	 * 获取默认数据库名称
	 */
	public String getDefaultCatalog() {
		return defaultCatalog;
	}

	public SqlDialect getDialect() {
		return dialect;
	}

	/**
	 * 获取函数。
	 * 
	 * @throws SQLException
	 */
	public List<String> getFunctions() throws SQLException {
		return this.queryMetaData(dialect
				.getQueryFunctions(getDefaultCatalog()));
	}

	/**
	 * 获取所有存储过程名称。
	 * 
	 * @throws SQLException
	 */
	public List<String> getProcedures() throws SQLException {
		return this.queryMetaData(dialect
				.getQueryProcedures(getDefaultCatalog()));
	}

	/**
	 * 获取{@link Connection}，并初始化。
	 * 
	 * @throws SQLException
	 */
	private Connection getPureConnection() throws SQLException {
		Connection conn = ds.getConnection();
		if (conn == null)
			throw new SQLException("当前状态获取Connection失败！");
		if (dialect == null) {
			dialect = Dialect.getDialect(
					conn.getMetaData().getDatabaseProductName())
					.getSqlDialect();
		}
		ai.getAndIncrement();
		return conn;
	}

	/**
	 * 获取表格。
	 * 
	 * @throws SQLException
	 */
	public List<String> getTables() throws SQLException {
		return this.getTables(new String[] { "TABLE" });
	}

	/**
	 * 获取表格。
	 * 
	 * @param types
	 *            表格类型列表。
	 * @throws SQLException
	 */
	public List<String> getTables(final String[] types) throws SQLException {
		return this.queryMetaData(dialect.getQueryTables(types,
				this.getDefaultCatalog()));
	}

	/**
	 * 初始化数据库，包括验证数据库以及自动创建数据库等。
	 * 
	 * @param conn
	 * @throws SQLException
	 */
	protected void initDao(Connection conn) throws SQLException {
	}

	protected <T, S> T query(AbstractQuery<T, S> exe) throws SQLException {
		return this.queryCore(exe);
	}

	/**
	 * 所有SQL执行的底层方法。 其实现了Connection的自动释放。
	 * 
	 * @param exe
	 * @throws SQLException
	 */
	protected <T> T queryCore(SqlQuery<T> exe) throws SQLException {
		Connection conn = this.getConnection();
		T t;
		try {
			conn.setAutoCommit(false);
			t = exe.query(conn);
			conn.commit();
		} catch (SQLException e) {
			conn.rollback();
			throw e;
		} finally {
			this.closeConnection(conn);
		}
		return t;
	}

	public <T> T queryMetaData(MetaDataResultHandler<T> exe)
			throws SQLException {
		return this.queryCore(new MetaDataQuery<T>(exe));
	}

	/**
	 * 设置默认数据库。 该设置将直接影响正在获取Connection的指令。
	 * 
	 * @param defaultCatalog
	 */
	protected void setDefaultCatalog(String defaultCatalog) {
		if (defaultCatalog == null
				|| !defaultCatalog.matches("[_a-zA-Z][_0-9a-zA-Z]{0,127}")) {
			if (defaultCatalog != null && defaultCatalog.equals(NULL))
				defaultCatalog = null;
			else
				throw new NullPointerException("数据库名称<" + defaultCatalog
						+ ">不符合要求!");
		} else
			this.defaultCatalog = defaultCatalog;
	}
}
