package me.icymint.libra.sage.core;

import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import javax.sql.DataSource;

import me.icymint.libra.sage.core.query.Execute;
import me.icymint.libra.sage.core.query.Query;
import me.icymint.libra.sage.core.result.ArrayListResult;
import me.icymint.libra.sage.core.result.ArrayResult;
import me.icymint.libra.sage.core.result.ResultHandler;
import me.icymint.libra.sage.core.result.ScalarListResult;
import me.icymint.libra.sage.core.result.ScalarResult;
import me.icymint.libra.sage.model.SqlVar;
import me.icymint.libra.sage.model.orm.DataPack;
import me.icymint.libra.sage.model.orm.Mappings;
import me.icymint.libra.sage.model.orm.PackException;
import me.icymint.libra.sage.model.orm.PackUtils;

/**
 * JdbcSupport的细化操作。其集成了很多基础的方法。 JdbcTemplate的替换版本，摆脱了Spring的依赖。
 * 
 * @author Daniel
 * @datetime 2012-11-9下午3:02:20
 * 
 */
public class JdbcUtils extends JdbcSupport {
	public static int[] getTypes(SqlVar[] vs) {
		if (vs == null)
			return new int[] {};
		int[] t = new int[vs.length];
		for (int i = 0; i < vs.length; i++) {
			t[i] = vs[i].getType().getJdbcType();
		}
		return t;
	}

	private final Mappings mps = new Mappings(new PackUtils());

	/**
	 * 
	 * @param ds
	 */
	public JdbcUtils(DataSource ds) {
		super(ds);
	}

	/**
	 * JdbcUtils初始化。
	 * 
	 * @param ds
	 *            数据源，其能自动识别数据源的信息，并返回方言版本。
	 * @param database
	 *            数据库名称，不可为空，数据源中设置的数据库将被替换。
	 */
	public JdbcUtils(DataSource ds, String database) {
		super(ds, database);
	}

	/**
	 * 执行单条无参数SQL语句
	 * 
	 * @param sql
	 * @throws SQLException
	 */
	public void execute(String sql) throws SQLException {
		executeBatch(sql, null);
	}

	/**
	 * 执行有参数，但没有返回值的Sql语句。
	 * <p>
	 * UPDATE、INSERT、DELETE语句，或者SQL DDL语句。
	 * 
	 * @param sql
	 * @param param
	 *            参数列表，可以没有参数。
	 * @throws SQLException
	 */
	public void execute(String sql, int[] types, Object... param)
			throws SQLException {
		executeBatch(sql, types, param);
	}

	/**
	 * 批量执行sql语句。 扩展了{@link #execute(String,int[], V...)}方法。
	 * 
	 * @param sql
	 *            sql语句
	 * @param param
	 *            sql参数
	 * @throws SQLException
	 */
	public void executeBatch(String sql, int[] types, Object[]... param)
			throws SQLException {
		super.execute(new Execute(sql, types, param));
	}

	public Mappings getMappings() {
		return this.mps;
	}

	/**
	 * 加载
	 * 
	 * @param sql
	 * @param cls
	 * @param vs
	 * @param key
	 * @param param
	 * @throws SQLException
	 */
	public <T> T load(String sql, Class<T> cls, SqlVar[] vs, SqlVar[] key,
			Object... param) throws SQLException {
		return loadList(sql, cls, vs, 1, 1, key, param).get(0);
	}

	public DataPack load(String sql, SqlVar[] vs, SqlVar[] key, Object... param)
			throws PackException, SQLException {
		return load(sql, DataPack.class, vs, key, param);
	}

	public <T> List<T> loadList(String sql, Class<T> cls, SqlVar[] vs,
			int offset, int maxRows, SqlVar[] key, Object... param)
			throws SQLException {
		List<T> list = new LinkedList<T>();
		for (Object[] objs : queryArrayList(sql, offset, maxRows,
				getTypes(key), param)) {
			try {
				list.add(mps.pack(objs, vs, cls));
			} catch (PackException e) {
				throw new SQLException("执行语句:" + sql + "\n"
						+ Arrays.toString(vs) + "\n" + Arrays.toString(objs)
						+ "\n" + sql + "\n" + Arrays.toString(param), e);
			}
		}
		return list;
	}

	public List<DataPack> loadList(String sql, SqlVar[] vs, int offset,
			int maxRows, SqlVar[] key, Object... param) throws SQLException {
		return loadList(sql, DataPack.class, vs, offset, maxRows, key, param);
	}

	/**
	 * 简单查询，无参数
	 * 
	 * @param sql
	 * @throws SQLException
	 */
	public <T> T query(String sql) throws SQLException {
		return query(sql, (int[]) null);
	}

	/**
	 * 查询具有一组参数的sql语句，并返回标量值。
	 * 
	 * @param sql
	 * @param param
	 * @throws SQLException
	 */
	public <T> T query(String sql, int[] types, Object... param)
			throws SQLException {
		return queryTemplate(sql, new ScalarResult<T>(), 1, types, param);
	}

	public Object[] queryArray(String sql) throws SQLException {
		return this.queryArray(sql, null);
	}

	/**
	 * 查询具有一组参数的sql语句，并返回一个数组值。
	 * 
	 * @param sql
	 * @param cls
	 *            返回数组的类。
	 * @param param
	 * @param types
	 * @throws SQLException
	 */
	public Object[] queryArray(String sql, int[] types, Object... param)
			throws SQLException {
		return queryTemplate(sql, new ArrayResult<Object>(Object.class), 1,
				types, param);
	}

	public List<Object[]> queryArrayList(String sql, int offset, int maxRows)
			throws SQLException {
		return this.queryArrayList(sql, offset, maxRows, null);
	}

	/**
	 * 查询具有一组String参数的sql语句，并返回一个数组的List。
	 * 
	 * @param sql
	 * @param cls
	 * @param param
	 * @param offset
	 * @param maxRows
	 * @param types
	 * @throws SQLException
	 */
	public List<Object[]> queryArrayList(String sql, int offset, int maxRows,
			int[] types, Object... param) throws SQLException {
		return queryTemplate(sql, new ArrayListResult<Object>(Object.class,
				offset), maxRows, types, param);
	}

	public <T> List<T> queryList(String sql, int offset, int maxRows)
			throws SQLException {
		return this.queryList(sql, offset, maxRows, null);
	}

	/**
	 * 查询具有一组String参数的sql语句，并返回一个List。
	 * 
	 * @param sql
	 * @param param
	 * @throws SQLException
	 */
	public <T> List<T> queryList(String sql, int offset, int maxRows,
			int[] types, Object... param) throws SQLException {
		return queryTemplate(sql, new ScalarListResult<T>(offset), maxRows,
				types, param);
	}

	/**
	 * 查询具有参数的sql语句，并返回一组值。
	 * 
	 * @param sql
	 * @param ph
	 * @param rh
	 * @param types
	 * @throws SQLException
	 */
	protected <T> T queryTemplate(String sql, ResultHandler<T> rh, int maxRows,
			int[] types, Object... params) throws SQLException {
		return query(new Query<T>(sql, rh, maxRows, types, params));
	}

	public <T> void update(String sql, SqlVar[] vs, T t) throws SQLException {
		try {
			execute(sql, getTypes(vs), mps.unpack(t, vs));
		} catch (PackException e) {
			throw new SQLException("属性" + t + "卸箱失败！", e);
		}
	}

	public <T> void updateBatch(String sql, SqlVar[] vs, Collection<T> coll)
			throws SQLException {
		Object[][] objs = new Object[coll.size()][];
		int i = 0;
		for (T t : coll) {
			try {
				objs[i++] = mps.unpack(t, vs);
			} catch (PackException e) {
				throw new SQLException("属性" + t + "卸箱失败！", e);
			}
		}
		executeBatch(sql, getTypes(vs), objs);
	}
}
