package me.icymint.libra.sage.core.dialect;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;

import me.icymint.libra.sage.core.result.MetaDataResultHandler;
import me.icymint.libra.sage.core.result.ScalarListResult;
import me.icymint.libra.sage.model.SqlFunction;
import me.icymint.libra.sage.model.SqlProcedure;
import me.icymint.libra.sage.model.SqlTable;
import me.icymint.libra.sage.model.SqlVar;
import me.icymint.libra.sage.model.SqlVarType;
import me.icymint.libra.sage.model.format.SqlFormatException;
import me.icymint.libra.sage.model.operator.SQLNotExistsException;

/**
 * DB2NT数据库。
 * 
 * @author Daniel
 * @datetime 2013-1-16下午5:24:13
 * 
 */
public class DB2NTDialect extends AbstractDialect {

	/**
	 * 函数结果查询。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-9下午2:09:42
	 * 
	 */
	private class Db2QueryFunctions extends ScalarListResult<String> implements
			MetaDataResultHandler<List<String>> {

		private final String defaultCatalog;

		Db2QueryFunctions(String defaultCatalog) {
			super(1, 1);
			this.defaultCatalog = defaultCatalog;
		}

		@Override
		public String handle(ResultSet rs, int collumnIndex)
				throws SQLException {
			return super.handle(rs, collumnIndex).toLowerCase();
		}

		@Override
		public ResultSet queryResultSet(DatabaseMetaData metadata)
				throws SQLException {
			return metadata
					.getConnection()
					.createStatement()
					.executeQuery(
							"SELECT ROUTINENAME FROM SYSIBM.SYSROUTINES WHERE ROUTINETYPE='F' AND "
									+ "CAST_FUNCTION='N' AND ORIGIN='Q' AND ROUTINESCHEMA='"
									+ defaultCatalog.toUpperCase() + "'");
		}
	}

	/**
	 * 存储过程结果查询。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-9下午2:10:10
	 * 
	 */
	private class Db2QueryProcedures extends ScalarListResult<String> implements
			MetaDataResultHandler<List<String>> {

		private final String defaultCatalog;

		private Db2QueryProcedures(String defaultCatalog) {
			super(1, 1);
			this.defaultCatalog = defaultCatalog;
		}

		@Override
		public String handle(ResultSet rs, int collumnIndex)
				throws SQLException {
			return super.handle(rs, collumnIndex).toLowerCase();
		}

		@Override
		public ResultSet queryResultSet(DatabaseMetaData metadata)
				throws SQLException {
			return metadata
					.getConnection()
					.createStatement()
					.executeQuery(
							"SELECT ROUTINENAME FROM SYSIBM.SYSROUTINES WHERE ROUTINETYPE='P' AND "
									+ "ORIGIN='Q' AND ROUTINESCHEMA='"
									+ defaultCatalog.toUpperCase() + "'");
		}
	}

	/**
	 * 查询表格信息。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-9上午11:54:10
	 * 
	 * @param <T>
	 */
	private class Db2QueryTables extends ScalarListResult<String> implements
			MetaDataResultHandler<List<String>> {
		private final String defaultCatalog;

		/**
		 * 构建查询表格的执行指令。
		 * 
		 * @param rh
		 *            结果处理类。
		 * @param types
		 *            表格类型，请参考
		 *            {@link DatabaseMetaData#getTables(String ,String ,String , String[])}
		 */
		Db2QueryTables(String[] types, String defaultCatalog) {
			super(1, 1);
			this.defaultCatalog = defaultCatalog;
		}

		@Override
		public String handle(ResultSet rs, int collumnIndex)
				throws SQLException {
			return super.handle(rs, collumnIndex).toLowerCase();
		}

		@Override
		public ResultSet queryResultSet(DatabaseMetaData metadata)
				throws SQLException {
			return metadata
					.getConnection()
					.createStatement()
					.executeQuery(
							"SELECT NAME FROM SYSIBM.SYSTABLES WHERE TYPE='T' AND CREATOR='"
									+ defaultCatalog.toUpperCase() + "'");
		}

	}

	protected DB2NTDialect(String name) {
		super(name);
	}

	@Override
	public String getCreateFunctionSql(SqlFunction sf)
			throws SqlFormatException {
		return null;
	}

	@Override
	public String getCreateProcedureSql(SqlProcedure sp)
			throws SqlFormatException {
		return null;
	}

	@Override
	public String getCreateTableSql(SqlTable st) throws SqlFormatException {
		String sql = "CREATE TABLE " + sqlPrefix() + st.toString() + "(\n";
		try {
			sql += this.var(st, 0);
			for (int i = 1; i < st.getVarSize(); i++) {
				sql += ls + this.var(st, i);
			}
			SqlVar[] key = st.getPrimaryKeys();
			if (key.length > 0) {
				sql += ls + "CONSTRAINT " + st.toString() + "_id PRIMARY KEY ("
						+ key[0].toString();
				for (int i = 1; i < key.length; i++) {
					sql += "," + key[i].toString();
				}
				sql += ")";
			}
			key = st.getForeignKeys();
			for (int i = 0; i < key.length; i++) {
				sql += ls + "FOREIGN KEY (" + key[i].toString()
						+ ") REFERENCES "
						+ st.getForeignTable(key[i].getName()).toString() + "("
						+ st.getForeignKey(key[i].getName()).toString() + ")";
			}
		} catch (SQLNotExistsException e) {
			throw new SqlFormatException("表格" + st.getName()
					+ "结构有问题！创建语句生成出错！", e);
		}
		String end = ")";
		return sql + end;
	}

	@Override
	public String getCurrentCatalog(Connection conn) throws SQLException {
		return this.query("VALUES CURRENT SCHEMA", conn);
	}

	@Override
	public String getDefaultCatalog(Connection conn) throws SQLException {
		return conn.getMetaData().getUserName();
	}

	@Override
	public String getDialectName() {
		return "db2nt";
	}

	@Override
	@Deprecated
	protected String getFunctionName(String function, String schema, int type) {
		return null;
	}

	@Override
	public String getInsert(SqlTable t) {
		String sql = "INSERT INTO " + this.sqlPrefix() + t.toString();
		if (t.isAutoIncrement()) {
			sql += " (";
			for (SqlVar v : t.getVars()) {
				if (!t.isKey(v.getName())) {
					sql += v.toString() + ",";
				}
			}
			sql = sql.replaceAll(",$", ")");
		}
		sql += " VALUES(?";
		int num = t.getVarSize();
		if (t.isAutoIncrement()) {
			num = num - t.getPrimaryKeys().length;
		}
		for (int i = 1; i < num; i++) {
			sql += ",?";
		}
		return sql + ")";
	}

	@Override
	@Deprecated
	protected String getProcedureName(String procedure, String schema, int type) {
		return null;
	}

	@Override
	public MetaDataResultHandler<List<String>> getQueryFunctions(
			String defaultCatalog) {
		return new Db2QueryFunctions(defaultCatalog);
	}

	@Override
	public MetaDataResultHandler<List<String>> getQueryProcedures(
			String defaultCatalog) {
		return new Db2QueryProcedures(defaultCatalog);
	}

	@Override
	public MetaDataResultHandler<List<String>> getQueryTables(String[] types,
			String defaultCatalog) {
		return new Db2QueryTables(types, defaultCatalog);
	}

	@Override
	protected void initialTypeMap(Map<SqlVarType, String> jdbc) {
		jdbc.put(SqlVarType.ARRAY, "BLOB");
		jdbc.put(SqlVarType.BINARY, "CHAR(n) FOR BIT DATA");
		jdbc.put(SqlVarType.BIT, "SMALLINT");
		jdbc.put(SqlVarType.BOOLEAN, "SMALLINT");
		jdbc.put(SqlVarType.FLOAT, "DOUBLE");
		jdbc.put(SqlVarType.JAVA_OBJECT, "BLOB");
		jdbc.put(SqlVarType.LONGVARBINARY, "LONG VARCHAR FOR BIT DATA");
		jdbc.put(SqlVarType.LONGVARCHAR, "LONG VARCHAR");
		jdbc.put(SqlVarType.NULL, "LONG VARCHAR FOR BIT DATA");
		jdbc.put(SqlVarType.NUMERIC, "DECIMAL");
		jdbc.put(SqlVarType.OTHER, "BLOB");
		jdbc.put(SqlVarType.STRUCT, "BLOB");
		jdbc.put(SqlVarType.TINYINT, "SMALLINT");
		jdbc.put(SqlVarType.VARBINARY, "VARCHAR(n) FOR BIT DATA");
	}

	@Override
	public boolean isSupportAutoCreateDatabase() {
		return false;
	}

	@Override
	public void setCurrentCatalog(Connection conn, String cat)
			throws SQLException {
		this.execute("SET SCHEMA " + cat, conn);
	}

	private String var(SqlTable st, int ind) throws SQLNotExistsException {
		SqlVar v = st.getVar(ind);
		String dv = v.getDefaultValue();
		return getVarName(v, st)
				+ " "
				+ v.getType().toString()
				+ (st.isNotNull(v.getName()) ? " NOT NULL" : "")
				+ (st.isUnique(v.getName()) ? " UNIQUE" : "")
				+ ((st.isAutoIncrement() && st.isKey(v.getName())) ? " GENERATED ALWAYS AS IDENTITY (START WITH 1, INCREMENT BY 1)"
						: "") + (dv == null ? "" : " DEFAULT '" + dv + "'");
	}
}
