package me.icymint.libra.sage.model;

import java.sql.SQLException;
import java.sql.Types;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import me.icymint.libra.sage.core.dialect.query.AdvancedQuery;
import me.icymint.libra.sage.core.dialect.query.AdvancedQueryFactory;
import me.icymint.libra.sage.core.dialect.query.AdvancedQueryImpl;
import me.icymint.libra.sage.core.dialect.query.SqlInnerFunction;
import me.icymint.libra.sage.model.object.SqlObjectType;
import me.icymint.libra.sage.model.object.TableManagerObject;
import me.icymint.libra.sage.model.operator.Curd;
import me.icymint.libra.sage.model.operator.ManagerObjects;
import me.icymint.libra.sage.model.operator.SQLExistsException;
import me.icymint.libra.sage.model.operator.SQLNotExistsException;
import me.icymint.libra.sage.model.operator.SqlManager;
import me.icymint.libra.sage.model.orm.DataPack;
import me.icymint.libra.sage.model.orm.PackException;
import me.icymint.libra.sage.model.orm.TableCurd;

/**
 * SQL表模型，用于表示数据库中的数据表元数据模型。 其可以进行JDBC操作和执行数据的增删查改。
 * 
 * @author Daniel
 * @datetime 2012-12-19下午2:33:53
 * 
 */
public class SqlTable extends TableManagerObject implements SqlManager, Curd {
	/**
	 * 封装了外键信息。
	 * 
	 * @author Daniel
	 * @datetime 2013-1-9上午9:10:36
	 * 
	 */
	private static class ForeignKey {
		private final SqlTable st;
		private final SqlVar key;

		ForeignKey(SqlTable st, SqlVar key) {
			this.st = st;
			this.key = key;
		}

		SqlVar getKey() {
			return key;
		}

		SqlTable getTable() {
			return st;
		}
	}

	private ManagerObjects<SqlVar> key = new ManagerObjects<SqlVar>(null,
			SqlVar.class);

	private ManagerObjects<SqlVar> notnull = new ManagerObjects<SqlVar>(null,
			SqlVar.class);
	private Map<String, ForeignKey> foreignkey = new LinkedHashMap<String, ForeignKey>();
	private ManagerObjects<SqlVar> unique = new ManagerObjects<SqlVar>(null,
			SqlVar.class);
	private boolean autoIncrement = false;

	private final TableCurd tc;

	protected SqlTable(SqlInfo si, String prefix, String name, boolean isMask,
			SqlVar... sv) {
		super(si, prefix, name, isMask, sv);
		this.tc = new TableCurd(this);
	}

	/**
	 * 创建高级查询。
	 * 
	 * @param cls
	 * @throws SQLException
	 */
	public <T> AdvancedQuery<T> advancedQuery(Class<T> cls) throws SQLException {
		return new AdvancedQueryImpl<T>(this, cls);
	}

	@Override
	public void create() throws SQLException {
		getJdbcUtils().execute(getDialect().getCreateTableSql(this));
	}

	@Override
	public void createIfNotExist() {
		try {
			this.create();
		} catch (SQLException e) {
		}
	}

	@Override
	public void delete() throws SQLException {
		getJdbcUtils().execute(getDialect().getDeleteTableSql(this));
	}

	@Override
	public <T> void delete(T t) throws SQLException {
		getCurdModel().delete(t);
	}

	@Override
	public <T> void deleteByKey(T t) throws SQLException {
		getCurdModel().deleteByKey(t);
	}

	/**
	 * 获取数据库中对应表格的记录数。
	 * 
	 * @throws SQLException
	 */
	public int getCounts() throws SQLException {
		return this
				.advancedQuery(Integer.class)
				.select(AdvancedQueryFactory.createQueryVar(this,
						SqlInnerFunction.COUNT)).commit().get(0);
	}

	protected TableCurd getCurdModel() {
		return tc;
	}

	/**
	 * 获取外键对应的外部表格的字段。
	 * 
	 * @param local
	 *            外键名称。
	 */
	public SqlVar getForeignKey(String local) {
		return foreignkey.get(local).getKey();
	}

	/**
	 * 获取所有外键列表。
	 */
	public SqlVar[] getForeignKeys() {
		SqlVar[] sv = new SqlVar[foreignkey.size()];
		int i = 0;
		for (String v : foreignkey.keySet()) {
			try {
				sv[i++] = getVar(v);
			} catch (SQLNotExistsException e) {
			}
		}
		// System.out.println(Arrays.toString(sv));
		return sv;
	}

	/**
	 * 获取外键对应的外部表格。
	 * 
	 * @param local
	 *            外键名称。
	 */
	public SqlTable getForeignTable(String local) {
		return foreignkey.get(local).getTable();
	}

	/**
	 * 获取数据库中该字段的最大取值。。
	 * 
	 * @param var
	 * @throws SQLException
	 */
	public int getMax(SqlVar var) throws SQLException {
		return this
				.advancedQuery(Integer.class)
				.select(AdvancedQueryFactory.createQueryVar(this, var,
						SqlInnerFunction.MAX)).commit().get(0);
	}

	/**
	 * 获取所有非空字段列表，包括主键和Unique键值。
	 */
	public SqlVar[] getNotNullVars() {
		return notnull.getAll();
	}

	/**
	 * 获取主键字段列表。
	 */
	public SqlVar[] getPrimaryKeys() {
		return key.getAll();
	}

	@Override
	protected SqlObjectType getSqlObjectType() {
		return SqlObjectType.Table;
	}

	/**
	 * 获取Unique字段列表。
	 */
	public SqlVar[] getUniqueVars() {
		return unique.getAll();
	}

	public <T> Object getValueOf(T t, SqlVar v) {
		try {
			return getJdbcUtils().getMappings().unpack(t, new SqlVar[] { v })[0];
		} catch (PackException e) {
			return null;
		}
	}

	@Override
	public <T> void insert(T t) throws SQLException {
		getCurdModel().insert(t);
	}

	@Override
	public <T> void insertAll(Collection<T> coll) throws SQLException {
		getCurdModel().insertAll(coll);
	}

	/**
	 * 只对单个主键生效。
	 */
	public boolean isAutoIncrement() {
		return this.getPrimaryKeys().length == 1
				&& getDialect().isAutoIncrement(
						this.getPrimaryKeys()[0].getType()) && autoIncrement;
	}

	/**
	 * 检查字段是否为外键。
	 * 
	 * @param local
	 */
	public boolean isForeignKey(String local) {
		return foreignkey.containsKey(local);
	}

	/**
	 * 检查字段是否为主键。
	 * 
	 * @param name
	 */
	public boolean isKey(String name) {
		try {
			key.get(name);
			return true;
		} catch (SQLNotExistsException e) {
			return false;
		}
	}

	/**
	 * 检查字段是否非空。
	 * 
	 * @param var
	 */
	public boolean isNotNull(String var) {
		try {
			notnull.get(var);
			return true;
		} catch (SQLNotExistsException e) {
			return false;
		}
	}

	/**
	 * 检查字段是否为Unique的。
	 * 
	 * @param var
	 */
	public boolean isUnique(String var) {
		try {
			unique.get(var);
			return true;
		} catch (SQLNotExistsException e) {
			return false;
		}
	}

	/**
	 * 表格关联性，返回值越小关联性低，0表示没有关联表格。
	 */
	public int level() {
		int level = 0;
		for (SqlTable t : this.getTables()) {
			level = Math.max(level, t.level() + 1);
		}
		return level;
	}

	@Override
	public List<DataPack> query() throws SQLException {
		return getCurdModel().query();
	}

	@Override
	public <T> List<T> query(Class<T> cls) throws SQLException {
		return getCurdModel().query(cls);
	}

	@Override
	public <T> List<T> query(Class<T> cls, int offset, int limit)
			throws SQLException {
		return getCurdModel().query(cls, offset, limit);
	}

	@Override
	public List<DataPack> query(int offset, int limit) throws SQLException {
		return getCurdModel().query(offset, limit);
	}

	@Override
	public <T> T queryByKey(Class<T> cls, Object... key) throws SQLException {
		return getCurdModel().queryByKey(cls, key);
	}

	@Override
	public DataPack queryByKey(Object... key) throws SQLException {
		return getCurdModel().queryByKey(key);
	}

	/**
	 * 设置自增，只适合单个字段作为主键，并且该字段类型为{@link Types#INTEGER}的
	 * 
	 * @param autoIncrement
	 */
	public void setAutoIncrement(boolean autoIncrement) {
		this.autoIncrement = autoIncrement;
	}

	/**
	 * 设置外键。
	 * 
	 * @param local
	 * @param t
	 * @param foreign
	 * @throws SQLException
	 */
	public void setForeignKey(String local, SqlTable t, String foreign)
			throws SQLException {
		SqlVar v;
		try {
			v = t.getVar(foreign);
		} catch (SQLNotExistsException e1) {
			throw new SQLException("表格" + t.getName() + "中不存在字段" + foreign, e1);
		}
		SqlVar n;
		try {
			n = this.getVar(local);
		} catch (SQLNotExistsException e1) {
			throw new SQLException("表格" + this.getName() + "中不存在字段" + local, e1);
		}
		if (v.getType() != n.getType())
			throw new SQLException("字段类型不匹配！" + v.getType() + ":" + n.getType());
		try {
			t.addTable(this);
		} catch (SQLExistsException e) {

		}
		setNotNull(local);
		// System.out.println(local+"-"+t+"."+foreign);
		foreignkey.put(local, new ForeignKey(t, v));
	}

	/**
	 * 设置非空字段。
	 * 
	 * @param v
	 * @throws SQLNotExistsException
	 */
	public void setNotNull(String v) throws SQLNotExistsException {
		try {
			notnull.add(getVar(v));
		} catch (SQLExistsException e) {
		}
	}

	/**
	 * 设置主键字段。
	 * 
	 * @param keys
	 * @throws SQLNotExistsException
	 */
	public void setPrimaryKeys(String... keys) throws SQLNotExistsException {
		for (String key : keys) {
			if (this.isUnique(key)) {
				logger.error("删除已经为Unique字段的主键字段" + key + "。");
				unique.del(getVar(key));
			}
			try {
				this.key.add(this.getVar(key));
			} catch (SQLExistsException e) {
				/**
				 * 如果主键已存在，则忽略。
				 */
			}
		}
	}

	/**
	 * 设置Unique字段，主键默认为Unique。 同时把该字段添加到非空字段中。
	 * 
	 * @param v
	 * @throws SQLNotExistsException
	 */
	public void setUniqueVar(String v) throws SQLNotExistsException {
		if (!this.isKey(v))
			try {
				unique.add(getVar(v));
				setNotNull(v);
			} catch (SQLExistsException e) {
			}
	}

	public <T> void setValueOf(T t, SqlVar v, Object value)
			throws PackException {
		getJdbcUtils().getMappings().setValueOf(t, v, value);
	}

	@Override
	public <T> T sync(T t) throws SQLException {
		return getCurdModel().sync(t);
	}

	@Override
	public <T> void update(T t) throws SQLException {
		getCurdModel().update(t);
	}
}
