package me.icymint.libra.sage.model.orm;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.Date;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;

import me.icymint.libra.sage.model.SqlFactory;
import me.icymint.libra.sage.model.SqlVar;

/**
 * 关系对象模型的自动映射。
 * 
 * @author Daniel
 * @datetime 2012-12-12下午2:20:16
 * 
 */
public class Mappings {
	public static String FETCH_BY_ID = "fetchById";

	public static Method findMethod(Class<?> cls, String method,
			Class<?>... parameterTypes) {
		try {
			return cls.getMethod(method, parameterTypes);
		} catch (NoSuchMethodException e) {
			return null;
		}
	}

	private final Map<Class<?>, Mapping<?>> mapping = new LinkedHashMap<Class<?>, Mapping<?>>();
	private final Map<Class<?>, Mapping<?>> bak = new LinkedHashMap<Class<?>, Mapping<?>>();
	private final PackUtils pu;
	private Map<Class<?>, Class<?>> prime = new LinkedHashMap<Class<?>, Class<?>>();

	{
		prime.put(String.class, String.class);
		prime.put(Integer.class, int.class);
		prime.put(Long.class, long.class);
		prime.put(Float.class, float.class);
		prime.put(Double.class, double.class);
		prime.put(Timestamp.class, long.class);
		prime.put(Date.class, long.class);
	}

	public Mappings(PackUtils pu) {
		this.pu = pu;
	}

	public Mappings(PackUtils pu, Collection<Mapping<?>> coll) {
		this(pu);
		for (Mapping<?> mp : coll)
			this.register(mp);
	}

	public Collection<Mapping<?>> getAll() {
		return mapping.values();
	}

	@SuppressWarnings("unchecked")
	public <T> Mapping<T> getMapping(Class<T> cls) throws MappingException {
		Mapping<T> m = (Mapping<T>) mapping.get(cls);
		if (m == null) {
			m = (Mapping<T>) bak.get(cls);
			if (m == null) {
				// System.out.println("类型" + cls.getName() +
				// "映射未注册！将使用默认映射实现。");
				m = SqlFactory.newMapping(cls);
				bak.put(cls, m);
			}
		}
		return m;
	}

	public PackUtils getPackUtils() {
		return pu;
	}

	public boolean isPrimary(Class<?> cls) {
		return prime.containsKey(cls);
	}

	/**
	 * 自动封装成{@link DataPack}对象。
	 * 
	 * @param obj
	 * @param v
	 * @throws PackException
	 */
	public DataPack pack(Object[] obj, SqlVar[] v) throws PackException {
		try {
			Constructor<DataPack> c = DataPack.class.getConstructor(
					SqlVar[].class, Object[].class);
			return c.newInstance(v, obj);
		} catch (SecurityException e) {
			throw new PackException(e);
		} catch (NoSuchMethodException e) {
			throw new PackException(e);
		} catch (IllegalArgumentException e) {
			throw new PackException(e);
		} catch (InstantiationException e) {
			throw new PackException(e);
		} catch (IllegalAccessException e) {
			throw new PackException(e);
		} catch (InvocationTargetException e) {
			throw new PackException(e);
		}
	}

	/**
	 * 根据对象属性自动装箱。
	 * 
	 * @param obj
	 * @param v
	 * @param cls
	 * @throws PackException
	 */
	@SuppressWarnings("unchecked")
	public <T> T pack(Object[] obj, SqlVar[] v, Class<T> cls)
			throws PackException {
		if (obj == null)
			return null;
		if (obj.length != v.length)
			throw new PackException("参数个数与取值个数不匹配！");
		if (cls == null || cls.equals(DataPack.class))
			return (T) pack(obj, v);
		if (this.isPrimary(cls)) {
			if (1 != obj.length)
				throw new PackException("参数个数与取值个数不匹配！");
			if (obj[0].getClass().equals(cls)) {
				return (T) obj[0];
			} else {
				Method m;
				try {
					m = cls.getMethod("valueOf", String.class);
					return (T) m.invoke(cls, String.valueOf(obj[0]));
				} catch (SecurityException e) {
					throw new PackException(e);
				} catch (NoSuchMethodException e) {
					throw new PackException(e);
				} catch (IllegalArgumentException e) {
					throw new PackException(e);
				} catch (IllegalAccessException e) {
					throw new PackException(e);
				} catch (InvocationTargetException e) {
					throw new PackException(e);
				}
			}
		} else {
			Mapping<T> mp;
			try {
				mp = getMapping(cls);
			} catch (MappingException e) {
				throw new PackException(e);
			}
			T t;
			try {
				t = mp.newInstance();
				for (int i = 0; i < v.length; i++) {
					setValueOf(t, mp.getSetterMethod(v[i]), obj[i]);
				}
			} catch (InstantiationException e) {
				throw new PackException(e);
			} catch (IllegalAccessException e) {
				throw new PackException(e);
			} catch (IllegalArgumentException e) {
				throw new PackException(e);
			} catch (NoSuchMethodException e) {
				throw new PackException(e);
			}
			return t;
		}
	}

	public <T> void register(Class<T> cls, String text) throws MappingException {
		register(SqlFactory.newMapping(cls, text));
	}

	public <T> void register(Mapping<T> mp) {
		mapping.put(mp.getDefinedClass(), mp);
	}

	protected <T> void setValueOf(T t, Method m, Object o) throws PackException {
		try {
			m.invoke(t, pu.pack(o, m.getParameterTypes()[0]));
		} catch (IllegalArgumentException e) {
			throw new PackException(e);
		} catch (IllegalAccessException e) {
			throw new PackException(e);
		} catch (InvocationTargetException e) {
			throw new PackException(e);
		}
	}

	public <T> void setValueOf(T t, SqlVar v, Object value)
			throws PackException {
		if (t != null) {
			try {
				@SuppressWarnings("unchecked")
				Mapping<T> mp = (Mapping<T>) this.getMapping(t.getClass());
				this.setValueOf(t, mp.getSetterMethod(v), value);
			} catch (MappingException e) {
				throw new PackException(e);
			} catch (NoSuchMethodException e) {
				throw new PackException(e);
			}
		} else {
			throw new PackException("对象为null", new NullPointerException());
		}
	}

	/**
	 * 对象属性获取并卸箱。
	 * 
	 * @param obj
	 * @param v
	 * @throws PackException
	 */
	public <T> Object[] unpack(T obj, SqlVar[] v) throws PackException {
		if (this.isPrimary(obj.getClass())) {
			if (v.length != 1)
				throw new PackException("参数个数与取值个数不匹配！");
			Object[] x = (Object[]) Array.newInstance(obj.getClass(), 1);
			x[0] = obj;
			return x;
		} else if (obj instanceof DataPack) {
			return ((DataPack) obj).getValues();
		} else {
			try {
				@SuppressWarnings("unchecked")
				Mapping<T> mp = (Mapping<T>) getMapping(obj.getClass());
				Object[] o = new Object[v.length];
				for (int i = 0; i < v.length; i++) {
					try {
						o[i] = pu.unpack(mp.getGetterMethod(v[i]).invoke(obj));
					} catch (IllegalArgumentException e) {
						throw new PackException(e);
					} catch (IllegalAccessException e) {
						throw new PackException(e);
					} catch (InvocationTargetException e) {
						throw new PackException(e);
					} catch (NoSuchMethodException e) {
						throw new PackException(e);
					}
				}
				return o;
			} catch (MappingException e) {
				throw new PackException(e);
			}
		}
	}
}
